#include <memory>
#include <iostream>
#include <string>

#include "smart_ptr.h"

using namespace std;

struct Node {
  shared_ptr<Node> next; // Node*
  string value;

  Node(shared_ptr<Node> next, string value) : next(next), value(value) {}
  Node() = default;

  ~Node() {
    cout << "destroyed object with value " << value << endl;
  }
};

struct Foo {
  SharedPtr<Foo> next;
  string value;

  ~Foo() {
    cout << "destroyed object with value " << value << endl;
  }
};

void TestSharedPtr() {
  // each ptr has ref count of 1
  SharedPtr<Foo> a = SharedPtr<Foo>::Make();
  a->value = "a";
  {
    SharedPtr<Foo> b = SharedPtr<Foo>::Make();
    b->value = "b";
    {
      SharedPtr<Foo> c = SharedPtr<Foo>::Make();
      c->value = "c";
      b->next = c; // copy the cell, increment the count
      c->next = a;
      // the heap graph: b -> c -> a -> NULL
      a->next = b;
      
    } // c is not destroyed until b is destroyed
  } // when b is destroyed, so is c, and ref count of a goes down to 1

  // finally, a is destroyed
}

void TestSharedPtr2() {
  // each ptr has ref count of 1
  SharedPtr<Foo> a = SharedPtr<Foo>::Make();
  a->value = "a";
  // set a to a new object
  a = SharedPtr<Foo>::Make();
  a->value = "a'";
}

Node * someFunctionInAnotherLibrary(shared_ptr<Node> p) {
  return p.get();
}

int main(int argc, char *argv[]) {
  //TestSharedPtr();
  //cout << "---" << endl;
  //TestSharedPtr2();
  
  // all 3 of these are shared_ptr<Node>
  // and, they all have a reference count of 1
  auto a = make_shared<Node>();
  auto b = make_shared<Node>();
  auto c = make_shared<Node>(a, "x");

  a->value = "y";
  b->value = "z";

  a->next = b; // ref count of z is now 2
  b->next = c; // ref count of x is now 2

  a->next = make_shared<Node>(nullptr, "u"); // ref count of z is 1

  weak_ptr d = b;
  Node * e = someFunctionInAnotherLibrary(b);
  b = nullptr;

  cout << "is d valid? " << (! d.expired()) << endl;

  cout << "value inside d: " << d.lock()->value << endl;

  cout << "end of main\n";
    // cout << a->value << b->value << c->value << endl;

  return 0;
}
